<?php
/**
 * This file implements the CommentGravatars plugin for Quam Plures
 *
 * This plugin displays a commentator's gravatar (Globally Recognized Avatar) if they've registered for a
 * free account with gravatar.com. If not, it displays one of three uniquely generated
 * avatars (identicons, wavatars or monsterids -- your choice).
 *
 * Designed as a no muss, no fuss way to add avatars to the comments of your blog without having
 * to bother with uploading files or doing any template coding.
 *
 * Works for registered users of your blog as well as visitors.'
 *
 * Quam Plures - {@link http://quamplures.net/}
 * Released under GNU GPL License - {@link http://quamplures.net/license.html}
 * @copyright (c) 2009 - 2011 by the Quam Plures developers - {@link http://quamplures.net/}
 *
 * @package plugins
 * @subpackage CommentGravatars
 */
if( !defined('QP_MAIN_INIT') ) die( 'Please, do not access this page directly.' );

/**
 * @package plugins
 * @subpackage CommentGravatars
 */
class comment_gravatars_plugin extends Plugin
{
	var $apply_rendering = 'lazy';
	var $author = 'Bush League Critic';
	var $author_url = 'http://www.bushleaguecritic.com/';
	var $code = 'comment_gravatars';
	var $group = 'comments';
	var $sub_group; // used if $group is 'other'
	var $number_of_installs = 1;
	var $priority = 50;
	var $version = '0.1';


	/**
	 * @see Plugin::PluginInit()
	 */
	function PluginInit( & $params )
	{
		$this->name = $this->T_('Comment Gravatars');
		$this->short_desc = $this->T_( 'Simple plugin to display avatars/gravatars for people leaving comments' );
		$this->long_desc = $this->T_( '<p>This plugin displays a commentator\'s gravatar (Globally Recognized Avatar) if they\'ve registered for a
 free account with gravatar.com. If not, it displays one of three uniquely generated avatars (identicons, wavatars or monsterids
 -- your choice).</p><p>Designed for a no muss, no fuss way to add avatars to the comments of your blog without having to bother
 with uploading files or doing any template coding.</p><p>Works for registered users of your blog as well as visitors.</p>' );
	}


	/**
	 * @see Plugin::GetDefaultSettings()
	 */
	function GetDefaultSettings( & $params )
	{
		$default_settings = array(
			'no_gravatar' => array(
				'label' => $this->T_('No Gravatar'),
				'type' => 'select',
				'defaultvalue' => 'identicon',
				'options' => array(
					'identicon' => $this->T_('Uniquely generated fractal glyph (Identicons)'),
					'monsterid' => $this->T_('Uniquely generated monster (Monsterids)'),
					'wavatar' => $this->T_('Uniquely generated face (Wavatars)'),
					// Not sure why anyone would want the same avatar for unknown users
					// but I'm leaving this last one as an option just in case.
					'default' => $this->T_('Gravatar default (Blue \'G\')'),
				),
				'note' => '<br />'.sprintf( $this->T_('What to display if commentator isn\'t registered at %s').'<br />',
					'<a href="http://www.gravatar.com" target="_blank">gravatar.com</a>' )
			),
			'test_start' => array(
				'label' => $this->T_('Avatar Preview'),
				'layout' => 'begin_fieldset',
			),
			'test' => array(
				'label' => $this->T_('Test email address'),
				'defaultvalue' => 'tester@example.com',
				'size' => 25,
				'note' => '<br />'.$this->T_('Enter an email address to preview the image.').'<br />',
			),
			'test_end' => array(
				'layout' => 'end_fieldset',
			),
		);
		return $default_settings;
	}


	/**
	 * This is what makes it all go.
	 */
	function display_gravatar( $params )
	{
		// Get the settings from the params and plugin settings
		$tmp_Blog = $params['Comment']->Item->get_Blog();
		$params = array_merge( array(
				'random' => $this->get_coll_setting( 'no_gravatar', $tmp_Blog ),
				'rating' => $this->get_coll_setting( 'gravatar_rating', $tmp_Blog ),
				'before' => '<div style="'.$this->get_coll_setting( 'gravatar_justify', $tmp_Blog ).'">',
				'after' => '</div>',
				'size' => $this->get_coll_setting( 'gravatar_size', $tmp_Blog ),
				'gravatar_title' => ( empty( $params['Comment']->author_User ) ? $params['Comment']->author : $params['Comment']->author_User->get_preferred_name() ),
				'gravatar_email' => ( empty( $params['Comment']->author_User ) ? $params['Comment']->author_email : $params['Comment']->author_User->email ),
			), $params );

		// Build the url to retrieve the correct gravatar from gravatar.com.
		$url = 'http://www.gravatar.com/avatar.php?gravatar_id='.md5( $params['gravatar_email'] );
		$url .= '&amp;rating='.urlencode( $params['rating'] );
		if ($params['random'] != 'default') {
			// If the email address isn't registered at gravatar.com;
			// Return one of three styles of uniquely generated 'identicons'.
			$url .= '&amp;default='.urlencode( $params['random'] );
		}

		// Build the html tags necessary to display the gravatar.
		$result = $params['before'];
		$result .=
		'<img '.
			'class="comment_gravatars_plugin_img" '.
			'src="'.$url.'" '.
			'width="'.$params['size'].'" '.
			'height="'.$params['size'].'" '.
			'title="'.format_to_output( $params['gravatar_title'], 'htmlattr' ).'" '.
			'alt="'.format_to_output( $params['gravatar_title'], 'htmlattr' ).'" '.
		'/>';
		$result .= $params['after'];

		// Display the gravatar!
		echo $result;

	}


	/**
	 * @see Plugin::TemplateBeginHtmlHead()
	 */
	function TemplateBeginHtmlHead()
	{
		// Add some styling to the avatar image.
		add_css_headline( 'img.comment_gravatars_plugin_img {
			margin:7px;
			padding:1px;
			border:1px solid #CCCCCC;
			background-color:white;
		}' );
	}


	/**
	 * @see Plugin::AdminEndHtmlHead()
	 */
	function AdminEndHtmlHead( & $params )
	{
		// Make sure the styling applies to the backoffice as well
		$this->TemplateBeginHtmlHead();
	}


	/**
	 * @see Plugin::PluginSettingsEditDisplayAfter()
	 */
	function PluginSettingsEditDisplayAfter( & $params )
	{
		$url = $url = 'http://www.gravatar.com/avatar.php?gravatar_id='.md5( $this->Settings->get( 'test' ) );
		$url .= $url .= '&amp;rating='.urlencode( 'g' );
		if ( $this->Settings->get( 'no_gravatar' ) != 'default' ) {
			$url .= '&amp;default='.urlencode( $this->Settings->get( 'no_gravatar' ) );
		}
		$result =
		'<img '.
			'class="comment_gravatars_plugin_img" '.
			'src="'.$url.'" '.
			'width="60" '.
			'height="60" '.
			'style="margin:7px;padding:1px;border:1px solid #CCCCCC;background-color:white;" ';
			'title="'.$this->Settings->get( 'no_gravatar' ).'" '.
			'alt="'.$this->Settings->get( 'no_gravatar' ).'" '.
		'/>';

		echo '<div class="label"><b>'.$this->T_('Preview').':</b></div>
			<div class="input">'.$result.'<div class="note">
			Click "Save (and review)" to see how changes affect the image.</div>';
	}


	/**
	 * @see Plugin::get_coll_setting_definitions()
	 */
	function get_coll_setting_definitions( & $params )
	{
		return array(
			'enable_avatars' => array(
				'label' => $this->T_('Enable Comment Avatars'),
				'type' => 'checkbox',
				'defaultvalue' => 1,
				'note' => sprintf( $this->T_('Check this to enable the %s plugin for this blog.'), $this->name ),
			),
			'no_gravatar' => array(
				'label' => $this->T_('No Gravatar'),
				'type' => 'select',
				'defaultvalue' => 'identicon',
				'options' => array(
					'identicon' => $this->T_('Uniquely generated fractal glyph (Identicons)'),
					'monsterid' => $this->T_('Uniquely generated monster (Monsterids)'),
					'wavatar' => $this->T_('Uniquely generated face (Wavatars)'),
					'default' => $this->T_('Gravatar default (Blue \'G\')'),
				),
				'note' => sprintf($this->T_('What to display if commentator isn\'t registered at %s'),
					'gravatar.com' ),
			),
			'gravatar_justify' => array(
				'label' => $this->T_('Display Avatar'),
				'type' => 'select',
				'defaultvalue' => 'float:left',
				'options' => array(
					'float:left' => $this->T_('Left of the comment'),
					'float:right' => $this->T_('Right of the comment'),
				),
				'note' => $this->T_('Place the gravatar to the left or right of the comment'),
			),
			'gravatar_size' => array(
				'label' => $this->T_('Avatar Size'),
				'type' => 'select',
				'defaultvalue' => '60',
				'options' => array(
					'40' => $this->T_('40x40 Pixels'),
					'60' => $this->T_('60x60 Pixels'),
					'80' => $this->T_('80x80 Pixels'),
				),
				'note' => $this->T_('What size gravatar image to display'),
			),
			'gravatar_rating' => array(
				'label' => $this->T_('Gravatar Rating'),
				'type' => 'select',
				'defaultvalue' => 'g',
				'options' => array(
					'g' => $this->T_('G-Rated gravatars'),
					'pg' => $this->T_('PG/G-Rated gravatars'),
					'r' => $this->T_('R/PG/G-Rated gravatars'),
					'x' => $this->T_('X/R/PG/G-Rated gravatars'),
				),
				'note' => $this->T_('Limit the explicitness of commentators gravatars'),
			),
		);
	}


	/**
	 * @see Plugin::FilterCommentContent()
	 */
	function FilterCommentContent( & $params )
	{
		if ( ! $this->get_coll_setting( 'enable_avatars', $params['Comment']->Item->get_Blog() ) )
		{
			return;
		}
		return $this->display_gravatar( $params );
	}

}

?>
